package com.tistory.devyongsik.moon.lucene.analysis.korean;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.document.Field.Index;
import org.apache.lucene.document.Field.Store;
import org.apache.lucene.index.CorruptIndexException;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.index.IndexWriter.MaxFieldLength;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Query;
import org.apache.lucene.search.ScoreDoc;
import org.apache.lucene.search.TermQuery;
import org.apache.lucene.search.TopDocCollector;
import org.apache.lucene.store.LockObtainFailedException;
import org.apache.lucene.store.RAMDirectory;

/**
 * 동의어 색인으로부터 하나의 Token에 대한 동의어를 가져온다.
 *
 * @author : 장용석, 2008. 11. 11., need4spd@google.com
 */

public class KoreanSynonymEngine implements SynonymEngine {
	Log logger = LogFactory.getLog(KoreanSynonymEngine.class);

	private static RAMDirectory directory;
	private static IndexSearcher searcher;
	private static SynonymEngine synonymEngineInstance = new KoreanSynonymEngine();
	private static final String SYONYM_DICTIONARY_NAME = "synonym.txt";
	
	//private AnalysisPropertiesReader reader = AnalysisPropertiesReader.getInstance();

	public static SynonymEngine getInstance() {
		return synonymEngineInstance;
	}

	private KoreanSynonymEngine() {
		if(logger.isInfoEnabled())
			logger.info("동의어 색인을 실시합니다.");
		
		createSynonymIndex();
		
		if(logger.isInfoEnabled())
			logger.info("동의어 색인 완료");
		
		try {
			searcher = new IndexSearcher(directory);
		} catch (CorruptIndexException e) {
			logger.error("동의어 색인에 대한 Searcher 생성 중 에러 발생함 : " + e);
			e.printStackTrace();
		} catch (IOException e) {
			logger.error("동의어 색인에 대한 Searcher 생성 중 에러 발생함 : " + e);
			e.printStackTrace();
		}
	}
	
	private void createSynonymIndex() {
		directory = new RAMDirectory();
		try {
			IndexWriter ramWriter = new IndexWriter(directory, new StandardAnalyzer(), true, MaxFieldLength.LIMITED);
			//InputStream inputStream = reader.getResourceStream(SYONYM_DICTIONARY_NAME);
			InputStream inputStream = getClass().getClassLoader().getResourceAsStream(SYONYM_DICTIONARY_NAME);
			
			BufferedReader in = null;
			in = new BufferedReader( new InputStreamReader(inputStream ,"utf-8"));

			String buffer = "";
			Field field;
			String fieldValue = "";
			Document doc = null;
			
			int recordCnt = 0;
			
			while( (buffer = in.readLine()) != null ) {
				String tmp[] = buffer.split(",");
				
				doc = new Document();

				for(int i = 0, size = tmp.length; i < size ; i++) {
					fieldValue  = tmp[i];
					field = new Field("syn",fieldValue,Store.YES,Index.NOT_ANALYZED);
		    		doc.add(field);
				}//end for
				recordCnt++;
				
				ramWriter.addDocument(doc);
			}
			ramWriter.optimize();
			ramWriter.close();
			
			if(logger.isInfoEnabled())
				logger.info("동의어 색인 단어 갯수 : " + recordCnt);
			
		} catch (CorruptIndexException e) {
			logger.error("동의어 색인 중 에러 발생함 : " + e);
			e.printStackTrace();
		} catch (LockObtainFailedException e) {
			logger.error("동의어 색인 중 에러 발생함 : " + e);
			e.printStackTrace();
		} catch (IOException e) {
			logger.error("동의어 색인 중 에러 발생함 : " + e);
			e.printStackTrace();
		}
	}
	
	public String[] getSynonyms(String word) throws IOException {
		ArrayList<String> synWordList = new ArrayList<String>();

		Query query = new TermQuery(new Term("syn",word));
		TopDocCollector collector = new TopDocCollector(10);
	    searcher.search(query, collector);
	    ScoreDoc[] hits = collector.topDocs().scoreDocs;

	    if(logger.isDebugEnabled()) {
    		logger.debug("대상 word : " + word);
    		//검색된 document는 하나이므로..
    		logger.debug("동의어 갯수 : " + hits.length);
    	}

	    for(int i = 0; i < hits.length; i++) {
	    	Document doc = searcher.doc(hits[i].doc);

	    	String[] values = doc.getValues("syn");

	    	for(int j = 0; j < values.length; j++) {
	    		if(logger.isDebugEnabled())
	    			logger.debug("대상 word : " + "["+word+"]" + " 추출된 동의어 : " + values[j]);

	    		if(!word.equals(values[j])) {
	    			synWordList.add(values[j]);
	    		}
	    	}
	    }
		return (String[]) synWordList.toArray(new String[0]);
	}
}